#ifndef HEADER_hostname
#define HEADER_hostname
/*
htop - generic/hostname.h
(C) 2021 htop dev team
Released under the GNU GPLv2+, see the COPYING file
in the source distribution for its full text.
*/

#include <stddef.h>


void Generic_hostname(char* buffer, size_t size);

#endif
