#ifndef HEADER_SystemdMeter
#define HEADER_SystemdMeter

/*
htop - SystemdMeter.h
(C) 2020 htop dev team
Released under the GNU GPLv2+, see the COPYING file
in the source distribution for its full text.
*/

#include "Meter.h"


extern const MeterClass SystemdMeter_class;

#endif /* HEADER_SystemdMeter */
